/**
 * Abstract:
 * Implement a modified version of the well-known Bakery Algorithm called
 * Black-White Bakery Algorithm.
 * Verify with the aid of LTL the following properties:
 * - Mutual Exclusion
 * - Progress
 * - lockout Freedom
 * - Deadlock (no invalid end-states)
 *
 * author: Patrick Trentin <patrick.trentin@unitn.it>
 *
 */
 
 /* GLOBAL RdA - SHARED TYPES AND VARIABLES DEFINITIONS */

// Number of Customers, should belong to the  range [0, 15]
// (More concurrently created processes are not allowed in spin)
#define NUM_CUSTOMERS 2
// Number of times each Customer goes back to the Backery, should belong to the  range [0, 15]
#define NUM_PURCHASES 15
// I'm more keen with caps-locked true/false values
#define TRUE 1
#define FALSE 0

// Definition of the available colour types.
// Note that the actual implementation of the algorithm is strictly dependant
// on the number and name of these types and may require further adjustment to work upon
// modifies this line.
mtype = {black, white};
// The sharedColour states the queue of customers currently being created. All
// of its customers must wait until the other queue (older) has been served.
// It's particularly important because by switching from one queue to the other
// we are ensuring our self that we won't have OVERFLOW by subsequently choosing
// greater and greater ticket numbers (Each time we switch we start from 1
// and the number of processes is strictly less than the SHORT_MAX).
mtype sharedColour = black;

// Ticket structure: each customer is assigned to a particular
// queue (colour) and has a turn based on his arrival (number).
typedef Ticket {
	int number ;	// If all the processes fall in the same queue and have different 
							// numbers, the range of this variable belongs to [0, 127], hence 
							// a byte is not sufficient ([0, 255])
	mtype colour;
	bool valid = FALSE;
};

// Boolean Vector, for each index n s.t. 0 <= n < NUM_CUSTOMERS, 
// choosing[n] == TRUE if process with _pid == n is picking up a 
// number in the meanwhile (FALSE otherwise).
bool choosingQueue[NUM_CUSTOMERS];

// Vector of Tickets, for each index n s.t 0 <= n < NUM_CUSTOMERS,
// assume that vectTickets[n] belongs to the process _pid and 
Ticket ticketsVector[NUM_CUSTOMERS];

/**
 * NOTE: the Black-White Bakery algorithm implementation *strongly* relies
 * on the assumption that each process accesses only to his portion
 * of the preceeding two vectors. Never *ever* change this behaviour
 */

// Number of Processes in Critical Section in a given moment,
// used for extra verification of Mutual Exclusion 
// through an assertion statement
int numProcInCritical = 0;


/* CUSTOMER PROCESS IMPLEMENTATION */

active [NUM_CUSTOMERS] proctype Customer () { 
	// Other selfexplainatory variables
	int index = 0;
	int cycle = 0;
	int max = 0;	
	do
	:: cycle < NUM_PURCHASES ->
		cycle++;
		/* BEGIN: TRIAL SECTION */
Trial:	{
			/* GET A TICKET (Colour + Number) */

			// We let the other processes know that we're choosing a colour by
			//  setting the boolean value to TRUE into the appropriate vector.
			choosingQueue[_pid] = TRUE;

			// The ticket colour should be the one of the most recent queue
			ticketsVector[_pid].colour = sharedColour;

			// Now we pick up a value that we *believe* to be the MAX of the 
			// chosen values for the same colour up to now. Note that because this 
			// section is *NOT* MUTUALLY EXCLUSIVE, it may be likely the case
			// that more than one process has the same number for the same colour.
			// It's only meant to provide a sort of FIFO access to a resource/critical section
			// which indeed is quite aleatory and dependant on the actual process scheduling.
			index = 0;
			max = 0;
			{
				do
				:: index < NUM_CUSTOMERS ->
					if
					:: ticketsVector[index].valid && ticketsVector[index].colour == ticketsVector[_pid].colour && ticketsVector[index].number > max ->
						max = ticketsVector[index].number;
					:: else ->
						skip;
					fi;
					index++;
				:: else -> break;
				od;
				ticketsVector[_pid].number = max + 1;
			}			
			// We're not choosing anymore!
			atomic {
				ticketsVector[_pid].valid = TRUE;
				choosingQueue[_pid] = FALSE;
			}

			/* WAIT UNTIL YOUR TURN */

			index = 0;
			do
			:: index < NUM_CUSTOMERS ->
				// If the process with _pid == index is choosing a number, he may or may be not gone there
				// before/concurrently/after us. Hence it's safer to wait until he has picked up a number.
				do
				:: ! (choosingQueue[index]) ->
					break;
				od;
				// Now, the process with _pid == index is not trying to get a number or will try to get it after
				// us, so he will be surely blocked later because i've been first in the ticket extraction than him.

				// For every process p, wait that he has gotten his turn in the critical section if
				// he has precedence. 
				do
				// If the ticket being holded by the other process is valid (not used), the two colours are different
				// and mine is not equal to the shared colour, then it means that i gained the ticket before him
				// and there's no reason to wait behind.
				:: ticketsVector[index].valid && ticketsVector[_pid].colour != ticketsVector[index].colour && ticketsVector[_pid].colour != sharedColour -> 
					break; 	// i'm in the colouredQueue currently being served.. 
						// no reason for waiting the other guy
				// If the ticket being holded by the other process is valid (not used), the two colours are equal
				// and mine has a number that is lower OR an equal but I have a lower PID, then i can skip waiting
				:: ticketsVector[index].valid && ticketsVector[_pid].colour == ticketsVector[index].colour && 
					(ticketsVector[_pid].number < ticketsVector[index].number || (ticketsVector[_pid].number == ticketsVector[index].number && _pid < index)) ->
					break;
				// If the ticket being holded by the other process is *NOT* valid (already used), then he had precedence
				// in respect to us but he HAS to get another ticket to access the Backery another time.
				// Hence we can skip waiting this guy
				:: ! ticketsVector[index].valid ->
					break;
				// Beware, you should never wait behind yourself
				:: _pid == index ->
					break;
				od;
				// Increase the FOR cycle guard value, to repeat every check for each process
				index++;
			:: else ->
				break;
			od;
		}
		/* END: TRIAL SECTION */

		/* BEGIN: CRITICAL SECTION */
Critical:	{
			// Very simple verification of mutual exclusion
			numProcInCritical++;
			assert(numProcInCritical == 1);
			printf("Hey, i'm process <%d> and i'm in critical section with the ticket: <%d, %d>.\n", _pid, ticketsVector[_pid].number, ticketsVector[_pid].colour);
			numProcInCritical--;

			// Before exiting i've to set the sharedColour
			if 
			:: ticketsVector[_pid].colour == black ->
				sharedColour = white;
			:: else ->
				sharedColour = black;
			fi;
			
			// Exit the critical section by resetting my number to 0
			atomic {
				ticketsVector[_pid].valid = FALSE;
				ticketsVector[_pid].number = 0;
			}
		}
		/* END: CRITICAL SECTION */
	:: else ->
		break;
	od;
}

/* LTL FORMULAs */

/**
 * Mutual Exclusion:
 * There's no reachable state in which more than 1 process is in the critical section.
 * In theory we could express it as
 * G ! ( For each process i,j s.t. i != j, Customer[i]@Critical && Customer[j]@Critical)
 */
ltl mutualExclusion2 { [] ! ((Customer[0]@Critical && Customer[1]@Critical)) }
 
ltl mutualExclusion3 { [] ! ((Customer[0]@Critical && Customer[1]@Critical) || (Customer[0]@Critical && Customer[2]@Critical) ||
								(Customer[1]@Critical && Customer[2]@Critical)) }
  /**
  * Progress:
  * we've to check that there isn't any execution in which we do not visit infinitively often a progress state (if we need it).
  * A progress state is a state in which we make some progress : I.E. critical section.
  * The general idea is: G(ORi Ti -> F ORi Ci), where i is a subscript that express universal quantification over the set of all states. 
  */
ltl progress2 { [] ( (Customer[0]@Trial || Customer[1]@Trial) -> <> (Customer[0]@Critical || Customer[1]@Critical) ) }

ltl progress3 { [] ( (Customer[0]@Trial || Customer[1]@Trial || Customer[2]@Trial) -> <> (Customer[0]@Critical || Customer[1]@Critical || Customer[2]@Critical) ) }

/**
 * Lockout-Freedom
 * In a fair path, if a process is in T, it eventually enters C.
 * i.e. FAIRRUN -> G (ANDi (Ti -> F Ci)), where i is a subscript that express universal quantification over the set of all states. 
 * Fairrun:
 * 	Let Ri be the true if the process with pid i is running. 
 * 	ANDi GF Ri, where i is a subscript that express universal quantification over the set of all states. 
 */
ltl lockoutFreedom2 { [] ( ( ([] <> (_last == 0)) && ([] <> (_last == 1)) ) -> 
						[] ( (Customer[0]@Trial -> <> Customer[0]@Critical) && (Customer[1]@Trial -> <> Customer[1]@Critical) ) ) }

//ltl lockoutFreedom3 { [] ( ( ([] <> (_last == 0)) && ([] <> (_last == 1)) && ([] <> (_last == 2)) ) -> 
//						[] ( (Customer[0]@Trial -> <> Customer[0]@Critical) && (Customer[1]@Trial -> <> Customer[1]@Critical) && (Customer[2]@Trial -> Customer[2]@Critical) ) ) }



